<?php
namespace MageDeveloper\Magepdf\Domain\Model;

abstract class AbstractModel implements \ArrayAccess
{
	/**
	 * Object attributes
	 *
	 * @var array
	 */
	protected $_data = array();

	/**
	 * Data changes flag (true after setData|unsetData call)
	 * @var $_hasDataChange bool
	 */
	protected $_hasDataChanges = false;

	/**
	 * Setter/Getter underscore transformation cache
	 *
	 * @var array
	 */
	protected static $_underscoreCache = array();

	/**
	 * __construct
	 * Constructor
	 */
	public function __construct()
	{
		$args = func_get_args();

		if (empty($args[0]))
		{
			$args[0] = array();
		}

		$this->_data = $this->_lowerArrayKeys($args[0]);
	}

	/**
	 * Lower all array keys if they are not complete uppercase
	 *
	 * @param array $array Array to lower
	 * @return array Lowered Array
	 */
	protected function _lowerArrayKeys($array)
	{
		$newArray = array();
		foreach($array as $_key=>$_val)
		{
			if (is_array($_val))
			{
				$newArray = array_merge($newArray, $this->_lowerArrayKeys($_val));
			}
			else
			{

				if (strtoupper($_key) == $_key)
				{
					$newKey = strtolower($_key);
					$newArray[$newKey] = $_val;
				}
				else
				{
					$newArray[$_key] = $_val;
				}

			}

		}

		return $newArray;
	}

	/**
	 * Get data change status
	 *
	 * @return bool
	 */
	public function hasDataChanges()
	{
		return $this->_hasDataChanges;
	}

	/**
	 * Retrieves data from the object
	 *
	 * If $key is empty will return all the data as an array
	 * Otherwise it will return value of the attribute specified by $key
	 *
	 * If $index is specified it will assume that attribute data is an array
	 * and retrieve corresponding member.
	 *
	 * @param string $key
	 * @param string|int $index
	 * @return mixed
	 */
	public function getData($key='', $index=null)
	{
		if (''===$key)
		{
			return $this->_data;
		}

		$default = null;

		// accept a/b/c as ['a']['b']['c']
		if (strpos($key,'/'))
		{
			$keyArr = explode('/', $key);
			$data = $this->_data;
			foreach ($keyArr as $i=>$k)
			{
				if ($k==='')
				{
					return $default;
				}

				if (is_array($data))
				{
					if (!isset($data[$k]))
					{
						return $default;
					}

					$data = $data[$k];

				}
				elseif ($data instanceof AbstractModel)
				{
					$data = $data->getData($k);
				}
				else
				{
					return $default;
				}

			}

			return $data;
		}

		// legacy functionality for $index
		if (isset($this->_data[$key]))
		{
			if (is_null($index))
			{
				return $this->_data[$key];
			}

			$value = $this->_data[$key];

			if (is_array($value))
			{
				//if (isset($value[$index]) && (!empty($value[$index]) || strlen($value[$index]) > 0)) {
				/**
				 * If we have any data, even if it empty - we should use it, anyway
				 */
				if (isset($value[$index]))
				{
					return $value[$index];
				}

				return null;

			}
			elseif (is_string($value))
			{
				$arr = explode("\n", $value);
				return (isset($arr[$index]) && (!empty($arr[$index]) || strlen($arr[$index]) > 0)) ? $arr[$index] : null;

			}
			elseif ($value instanceof AbstractModel)
			{
				return $value->getData($index);
			}

			return $default;
		}

		if (property_exists($this, $key))
		{
			return $this->{$key};
		}

		return $default;
	}

	/**
	 * Overwrite data in the object.
	 *
	 * $key can be string or array.
	 * If $key is string, the attribute value will be overwritten by $value
	 *
	 * If $key is an array, it will overwrite all the data in the object.
	 *
	 * @param string|array $key
	 * @param mixed $value
	 * @return AbstractModel
	 */
	public function setData($key, $value=null)
	{
		$this->_hasDataChanges = true;

		if (property_exists($this, $key))
		{
			$this->{$key} = $value;
		}

		if(is_array($key))
		{
			$this->_data = $key;
		}
		else
		{
			$this->_data[$key] = $value;
		}

		return $this;
	}

	/**
	 * If $key is empty, checks whether there's any data in the object
	 * Otherwise checks if the specified attribute is set.
	 *
	 * @param string $key
	 * @return boolean
	 */
	public function hasData($key='')
	{
		if (empty($key) || !is_string($key))
		{
			return !empty($this->_data);
		}

		return array_key_exists($key, $this->_data);
	}

	/**
	 * Add data to the object.
	 *
	 * Retains previous data in the object.
	 *
	 * @param array $arr
	 * @return AbstractModel
	 */
	public function addData(array $arr)
	{
		foreach($arr as $index=>$value)
		{

			if (strtoupper($index) == $index)
			{
				$index = strtolower($index);
			}

			$this->setData($index, $value);
		}

		return $this;
	}

	/**
	 * Unset data from the object.
	 *
	 * $key can be a string only. Array will be ignored.
	 *
	 * @param string $key
	 * @return AbstractModel
	 */
	public function unsetData($key=null)
	{
		$this->_hasDataChanges = true;

		if (is_null($key))
		{
			$this->_data = array();
		}
		else
		{
			unset($this->_data[$key]);
		}

		return $this;
	}

	/**
	 * Set/Get attribute wrapper
	 *
	 * @param   string $method
	 * @param   array $args
	 * @return mixed
	 * @throws \Exception
	 */
	public function __call($method, $args)
	{
		switch (substr($method, 0, 3))
		{
			case 'get' :
				$key = $this->_underscore(substr($method,3));
				$data = $this->getData($key, isset($args[0]) ? $args[0] : null);
				return $data;

			case 'set' :
				$key = $this->_underscore(substr($method,3));
				$result = $this->setData($key, isset($args[0]) ? $args[0] : null);
				return $result;

			case 'uns' :
				$key = $this->_underscore(substr($method,3));
				$result = $this->unsetData($key);
				return $result;

			case 'has' :
				$key = $this->_underscore(substr($method,3));
				return isset($this->_data[$key]);
		}

		throw new \Exception("Invalid method ".get_class($this)."::".$method."(".print_r($args,1).")");
	}

	/**
	 * Converts field names for setters and geters
	 *
	 * $this->setMyField($value) === $this->setData('my_field', $value)
	 * Uses cache to eliminate unneccessary preg_replace
	 *
	 * @param string $name
	 * @return string
	 */
	protected function _underscore($name)
	{
		if (isset(self::$_underscoreCache[$name]))
		{
			return self::$_underscoreCache[$name];
		}

		$result = strtolower(preg_replace('/(.)([A-Z])/', "$1_$2", $name));
		self::$_underscoreCache[$name] = $result;
		return $result;
	}

	/**
	 * checks whether the object is empty
	 *
	 * @return boolean
	 */
	public function isEmpty()
	{
		if (empty($this->_data))
		{
			return true;
		}

		return false;
	}

	/**
	 * Implementation of ArrayAccess::offsetSet()
	 *
	 * @link http://www.php.net/manual/en/arrayaccess.offsetset.php
	 * @param string $offset
	 * @param mixed $value
	 */
	public function offsetSet($offset, $value)
	{
		$this->_data[$offset] = $value;
	}

	/**
	 * Implementation of ArrayAccess::offsetExists()
	 *
	 * @link http://www.php.net/manual/en/arrayaccess.offsetexists.php
	 * @param string $offset
	 * @return boolean
	 */
	public function offsetExists($offset)
	{
		return isset($this->_data[$offset]);
	}

	/**
	 * Implementation of ArrayAccess::offsetUnset()
	 *
	 * @link http://www.php.net/manual/en/arrayaccess.offsetunset.php
	 * @param string $offset
	 */
	public function offsetUnset($offset)
	{
		unset($this->_data[$offset]);
	}

	/**
	 * Implementation of ArrayAccess::offsetGet()
	 *
	 * @link http://www.php.net/manual/en/arrayaccess.offsetget.php
	 * @param string $offset
	 * @return mixed
	 */
	public function offsetGet($offset)
	{
		return isset($this->_data[$offset]) ? $this->_data[$offset] : null;
	}
}